package {
	import org.flixel.FlxG;
	import org.flixel.FlxSprite;
	import org.flixel.FlxState;

	public class PlayState extends FlxState {
		/*
		 * Ship View
		 */
		public var sv:SpaceView;
		/*
		 * Internal View
		 */
		public var isv:InternalShipView;
		/*
		 * Script object
		 */
		private var _script:Script;

		/*
		 * There is no level 0
		 */
		private var _levelClasses:Array = [null, LevelOne, LevelTwo, LevelThree, LevelBezTest];

		/*
		 * Timer
		 */
		private var _fadeTimer:Number;
		/*
		 * Duration
		 */
		private var _fadeDuration:Number;
		/*
		 * In or Out
		 */
		private var _fadeIn:Boolean;
		/*
		 * Action to perform after fade finished
		 */
		private var _fadeAction:Function;
		/*
		 * Whether or not everything else is paused while fading
		 */
		private var _fadePause:Boolean;
		/*
		 * Helper sprite representing the fade mask
		 */
		private var _fadeMask:FlxSprite;

		override public function create():void {
			// Cinematics / Script
			_script = new Script();
			
			// Internal and Space Views			
			sv = null;
			isv = new InternalShipView();
			add(isv);
			
			// Start at 0 (which doesn't exist)
			FlxG.level = 0;
			// Load level 1
			loadInteriorLevel(1);
			
			// Fade logic
			_fadeMask = new FlxSprite(0, 0);
			_fadeMask.scrollFactor.x = _fadeMask.scrollFactor.y = 0;
			_fadeMask.createGraphic(FlxG.width, FlxG.height, 0xff000000);
			fadeIn(0.5, false);
		}

		override public function update():void {
			// Fade Logic
			if(_fadeDuration > 0) {
				if(_fadeTimer < 0) {
					// Done fading
					_fadeDuration = 0;
					if(_fadeAction != null)
						_fadeAction();
				} else {
					// Fade
					_fadeTimer -= FlxG.elapsed;
					if(_fadeIn) {
						_fadeMask.alpha = _fadeTimer / _fadeDuration;
					} else {
						_fadeMask.alpha = 1 - _fadeTimer / _fadeDuration;
					}
				
					// Don't update anything while fading if fadePause is on
					if(_fadePause)
					return;
				}
			}

			// Restart with escape key
			if(FlxG.keys.justPressed("ESCAPE"))
				FlxG.state = new MenuState();
			
			super.update();
		}

		/*
		 * Load specified level
		 */
		private function loadInteriorLevel(level:int, escape:Boolean = false):void {
			// Return to menu if all levels beaten
			if(level >= _levelClasses.length) {
				FlxG.state = new MenuState();
				return;
			}
			
			// Don't load cinematics if the player lost the last level
			if(level > FlxG.level) {
				// Load cinematics from state machine here, but only if level increased
				isv.queueupCinematics(_script.getCinematics(isv.getPositionTable()));
				if(level == 1) {
					isv.playCinematics(false);
				} else {
					isv.playCinematics(true);
				}
			} else {
				// Defeat logic here
				if(!escape) {
					// Only do defeat logic here which relies on a failed or non-escape
				}
			}
			
			FlxG.level = level;
			
			if(sv)
				sv.visible = sv.active = false;
			isv.visible = isv.active = true;
			isv.switchTo();
		}

		private function loadExteriorLevel():void {
			// Get upgrade levels
			var level:int = FlxG.level;
			var upgrades:Array = new Array();
			var posTable:Object = isv.getPositionTable();
			
			upgrades[Script.NAV] = 0;
			upgrades[Script.WEP] = 0;
			upgrades[Script.SHIELD] = 0;
			
			for each (var value:* in posTable) {
				upgrades[value]++;
			}
			
			FlxG.levels[level] = new _levelClasses[level](upgrades); // Get upgrade values from isv instead

			if(sv) {
				sv = SpaceView(defaultGroup.replace(sv, FlxG.levels[level]));
			} else {
				sv = SpaceView(add(FlxG.levels[level]));
			}
		}

		/*
		 * Called when the player has won the level in space view
		 * Destroy old level here
		 */
		public function victory():void {
			fadeOut(0.5, true, victoryComplete);
		}

		private function victoryComplete():void {
			// Destroy old level
			SpaceView(FlxG.levels[FlxG.level]).destroy();
			FlxG.levels[FlxG.level] = null;
			// Load next level
			loadInteriorLevel(FlxG.level + 1);
			
			fadeIn(0.5, true);
		}

		/*
		 * Called when the player has lost the level in space view
		 * Destroy old level here
		 */
		public function defeat(escape:Boolean):void {
			if(escape)
				fadeOut(0.5, true, escapeComplete);
			else
				fadeOut(0.5, true, defeatComplete);
		}

		private function escapeComplete():void {
			// Destroy old level
			SpaceView(FlxG.levels[FlxG.level]).destroy();
			FlxG.levels[FlxG.level] = null;
			// Re-load the level
			loadInteriorLevel(FlxG.level, true);
			
			fadeIn(0.5, true);			
		}

		private function defeatComplete():void {
			// Destroy old level
			SpaceView(FlxG.levels[FlxG.level]).destroy();
			FlxG.levels[FlxG.level] = null;
			// Re-load the level
			loadInteriorLevel(FlxG.level);
			
			fadeIn(0.5, true);			
		}

		/*
		 * Called by internal ship view to indicate the player has chosen to deploy
		 */
		public function deploy():void {
			fadeOut(0.5, true, deployComplete);
		}

		private function deployComplete():void {
			loadExteriorLevel();
			sv.visible = sv.active = true;
			isv.visible = isv.active = false;
			sv.switchTo();
			
			fadeIn(0.5, true);
		}

		private function fadeIn(duration:Number, pause:Boolean, action:Function = null):void {
			_fadeTimer = _fadeDuration = duration;
			_fadeIn = true;
			_fadePause = pause;
			_fadeAction = action;
		}

		private function fadeOut(duration:Number, pause:Boolean, action:Function = null):void {
			_fadeTimer = _fadeDuration = duration;
			_fadeIn = false;
			_fadePause = pause;
			_fadeAction = action;
		}

		override public function postProcess():void {
			if(_fadeDuration > 0) {
				_fadeMask.render();
			}
		}
	}
}
